
;====================
; DXLFVFO project
; DDS synthesizer up to 500 kHz
; (C) Pavel Vachal,OK1DX 2014
; version 1.0 AT Mega 8
; 
; HAMWARE - free for Amateur radio community
; commercial use only with author's approval
;====================
; history: initial release 1.0

.NOLIST

;*********  FUNCTION LIST  *********
; LOOP	- main program loop, started after initialisation
; DDSFR - calculates DDS phase increment from freqency [Hz]
; UFRQ - add/subtract freq step to current freq, check ranges
; FQ2ASCII - convert binary freq into ASCII string
; PFRQ - print ASCII freq to LCD, set cursor position

; LOOP - main program loop. Handling of buttons (buttons functions) and freq generation
; LCDC - send command to LCD
; LCDD - send data to LCD
; TEXT - send fixed text from EEPROM to LCD
; WAIT - time delay, up to 25.6 milisec (unit 100 microsec)


;******** INTERRUPT HANDLERS *******
; RESET - entry point
; TIM0_OVF - timer 0 overflow: handling of button autorepeat function
; TIM1_OVF - used for autosave to EEPROM timer


;******** FUSES ********************
; ATmega8 needs to program fuses to operates correctly with external Xtal resonator
; SUT1 = 1, SUT0 = 1, CKSEL3 = 1, CKSEL2 = 1, CKSEL1 = 1, CKSEL0 = 1
; higher fuses byte = 11111111
; fuses programming must be done separately (not performed during code programming)!!!!

.include	"m8def.inc"
.LIST

; port initialisation
; DDR 1=output, 0=input
; PORT when input 1=pullup 0=tri state
; Xtal freq = 16 MHz
.EQU	_DDRB	=0x0	; PB0-PB5 inputs
.EQU	_PORTB	=0x3B	; PB0,1,3,4,5 pullup, PB2 normal - PB0=CW key, PB1= , PB2=buttons
	.EQU KBDI	=2		; buttons input bit

.EQU	_DDRC	=0x3F	; PC0-5 output - display, PC2-5 - buttons
.EQU	_PORTC	=0x0
.EQU	_DDRD	=0xFF	; PD all outputs - R-2R network
.EQU	_PORTD	=0x80

.EQU	TIMRUN	= 0x5	; when sent to TCCR1B value: timer run, internal clock divided by 1024

; registers
// PHASE ACCUMULATOR, 24bit
.DEF	ACU1	=R1		; Phase accu, LSB
.DEF	ACU2	=R2		; 
.DEF	ACU3	=R26	; Phase accu, MSB. XL as well
// PHASE INCREMENT, 24 bit
.DEF	PHI1	=R3		; LSB
.DEF	PHI2	=R4
.DEF	PHI3	=R5		; MSB
.DEF	XX1		=R6		; temp storage, used for calculation
.DEF	XX2		=R7
.DEF	XX3		=R8
.DEF    RES1	=R9		; temp storage, used for calculation
.DEF	RES2	=R10
.DEF	RES3	=R11
.DEF	CTR		=R12	; loop counter
.DEF	ITIME	=R13	; timer for keyboard, decremented by IRQ every 10 msec
.DEF	SCTR	=R14	; timer for autosave freq. Decremented by TIM1
.DEF	ISREG	=R15	; temp SREG storage during processing of INT handler

.DEF	TMP1	=R16	; general temp regs used in main loop
.DEF	TMP2	=R17
.DEF	TMP3	=R18
.DEF	ITMP1	=R19	; general temp regs used in INT handlers
.DEF	ITMP2	=R20
.DEF	STREG	=R21	; status register
	.EQU	LCDR	=0	; request for LCD refresh
	.EQU	TOUT	=1	; timer timeout (button autorepeat)

.DSEG
; static RAM registers, max 1kB

EC1:; values initialized from EEPROM - begin eeprom copy

FRQN:	.BYTE	3	; freq in Hz initial value from EEPROM
FR1:	.BYTE	3	; minimum freq [Hz]
FR2:	.BYTE	3	; maximum freq [Hz]
FRCLK:	.BYTE	3	; freq DDS clock (XO / 10) from EEPROM

EC2:; values initialized from EEPROM - end eeprom copy

FRQA:	.BYTE	6	; freq in Hz as ASCII "123456"
KEY_ST:	.BYTE	1	; keyboard status b7=key is pressed, b0=key0, b1=key1, b2=key2, b3=key3
	.EQU PRESS	=7
	.EQU KEY0	=0
	.EQU KEY1	=1
	.EQU KEY2	=2
	.EQU KEY3	=3

CPSN:	.BYTE 1	; cursor position (0= 1Hz, 1= 10 Hz,... 3= 1 kHz, 5= 100 kHz)
BUF3B:	.BYTE 3	; buffer, used to convertion Freq to ASCII

.ORG 0x100
SINTBL:	.BYTE	256	; must start at address 0xXX00 !!!

; parameters
; LCD
.EQU	LCDWT	=70	; timing const 150 mikrosec
.EQU	LCDW1	=150	; time delay 15 msec
.EQU	LCDW2	=40		; time delay 4 msec 
; KEYBOARD / times (unit 16.384 msec)
.EQU	ARPT1	=61	; start of autorepeat (1 sec)
.EQU	ARPT2	=12		; running autorepeat (200 msec)
.EQU	KEYF	=6		; filter for key off

.CSEG
.ORG	0

; interrupt vector table	
	rjmp	RESET	;1	Reset Handler
	reti 		;2 IRQ0 Handler
	reti		;3 IRQ1 Handler
	reti		;4 TIMER2_COMP
	reti		;5 TIMER2_OVF
	reti		;6 rjmp TIM1_CAPT ; Timer1 Capture Handler
	reti		;7 rjmp TIM1_COMPA ; Timer1 compare Handler
	reti		;8 rjmp TIM1_COMPA ; Timer1 compare Handler
	rjmp	TIM1_OVF	;9 Timer1 Overflow Handler
	rjmp	TIM0_OVF	;10 Timer0 Overflow Handler
	reti		;11 SPI Transfer Complete Handler
	reti 		;12 UART RX Complete Handler
	reti 		;13 UDR Empty Handler
	reti		;14 rjmp UART_TXC ; UART TX Complete Handler
	reti		;15 ADC ; ADC Conversion Complete Interrupt Handler
	reti		;16 EEPROM Ready Handler
	reti		;17 Analog Comparator Handler
	reti		;18 TWI
	reti		;19 SPM_RDY
	
RESET:
; initiate stack pointer
	ldi	TMP1,0x5F		; end of RAM of Mega8
	out	SPL,TMP1		; stack
	ldi	TMP1,4
	out	SPH,TMP1

	ldi	TMP1,_DDRB	; port mode ini
	out	DDRB,TMP1
	ldi	TMP1,_DDRC
	out	DDRC,TMP1
	ldi	TMP1,_DDRD
	out	DDRD,TMP1
	ldi	TMP1,_PORTB
	out	PORTB,TMP1
	ldi	TMP1,_PORTC
	out	PORTC,TMP1
	ldi	TMP1,_PORTD
	out	PORTD,TMP1
	
; register INI
	clr TMP1
	sts KEY_ST, TMP1
	ldi TMP1,2
	sts CPSN, TMP1		; pointing to 100 Hz position
	ldi TMP1,0xFF
	mov SCTR,TMP1		; save counter waiting

; LCD MODULE INITIALIZATION
	ldi	TMP1,150
	rcall	WAIT		; wait 15 msec
	ldi	TMP1,0x30
	rcall	LCDX		; 8 bit mode (+ blabla)
	ldi	TMP1,40
	rcall	WAIT		; wait 4 msec
	ldi	TMP1,0x30
	rcall	LCDX		; 8 bit mode (+ blabla)
	rcall	MOM
	ldi	TMP1,0x20
	rcall	LCDX		; 4 bit mode (+ blabla)
	rcall	MOM
	ldi	TMP1,0x28
	rcall	LCDC		; 4 bits, 2 lines, 5x7 dots
	ldi	TMP1,0x08
	rcall	LCDC		; display OFF
	ldi TMP1,0x0E
	rcall	LCDC		; display ON, cursor ON, no blink
	ldi TMP1,0x06
	rcall	LCDC		; cursor move increment, no display shift
	ldi	TMP1,0x01	
	rcall	LCDC		; clear display
	ldi TMP1,LCDW2
	rcall	WAIT		; wait 4 msec
	
	ldi	TMP1,VERS
	rcall	TEXT		; LFDDS version OK1DX

	ldi TMP1,TXT1		; Frq:
	rcall	 TEXT

	ldi TMP1,TXT2		; kHz
	rcall	 TEXT

	
; copy sin table from EEPROM into SRAM
	ldi XL,LOW(SINTBL)	; destination
	ldi XH,HIGH(SINTBL)
	ldi	YL,LOW(EESIN)	; source
	ldi YH,HIGH(EESIN)
	clr	TMP3			; length=256 bytes
	rcall EE_BLK

; copy initial parameters from EEPROM into SRAM
	ldi XL,LOW(EC1)	; destination
	ldi XH,HIGH(EC1)
	ldi	YL,LOW(EEFRQ)	; source
	ldi YH,HIGH(EEFRQ)
	ldi	TMP3, EC2-EC1			; length of SRAM area
	rcall EE_BLK
 	
; TIMER 0 initialisation
	ldi	TMP1,5
	out	TCCR0,TMP1		; T0 8bit  internal clock 1:1024

	clr TMP1
	out TCCR1B,TMP1		; T1 16bit stopped
	
	ldi	TMP1,0x05		; TOIE0 = 1, TOIE1 = 1
	out	TIMSK,TMP1		; enable TIMER0 + TIMER1 overflow interrupts

	rcall DDSFR
	rcall FQ2ASCII
	rcall PFRQ

;********** MAIN PROGRAM LOOP *********
;it waits for events signalised by interrupt rutines
LOOP:
		sei		; enable interrupt (again)

;***** KEYBOARD HANDLING ****************
		lds	TMP1,KEY_ST
		sbrc	TMP1,PRESS	; skip, if key is not pressed
		rjmp	KBD1

		ldi	TMP1,1<<2	; test button KEY0
		out	PORTC,TMP1
		rcall MOM
		sbis	PINB,KBDI
		rjmp	KBD2	; key not pressed, try another key
		ldi	TMP1,1<<KEY0 | 1<<PRESS
		rjmp	KBD3

KBD2:	ldi	TMP1,1<<3	; test button KEY1
		out	PORTC,TMP1
		rcall MOM
		sbis	PINB,KBDI
		rjmp	KBD4	; key not pressed, try another key
		ldi	TMP1,1<<KEY1 | 1<<PRESS
		rjmp	KBD3

KBD4:	ldi	TMP1,1<<4	; test button KEY2
		out	PORTC,TMP1
		rcall MOM
		sbis	PINB,KBDI
		rjmp	KBD5	; key not pressed, try another key
		ldi	TMP1,1<<KEY2 | 1<<PRESS
		rjmp	KBD3

KBD5:	ldi	TMP1,1<<5	; test button KEY3
		out	PORTC,TMP1
		rcall MOM
		sbis	PINB,KBDI
		rjmp	KBDEND	; no key is pressed
		ldi	TMP1,1<<KEY3 | 1<<PRESS
KBD3:	sts KEY_ST, TMP1
		ldi TMP1, ARPT1	; initialize for 1000 msec wait (start keyboard autorepeat)
		mov ITIME,TMP1
		rjmp KBDAK

KBD1:	mov TMP1,ITIME
		cpi TMP1,ARPT1-KEYF
		brge	KBD6
		ldi TMP1, 0x3C	; test all keys together
		out PORTC,TMP1
		rcall MOM
		sbic	PINB,KBDI	; skip if nothing pressed
		rjmp	KBD6
		lds	TMP1,KEY_ST
		cbr TMP1,1<<PRESS
		sts KEY_ST,TMP1	; reset PRESS bit
		rjmp KBDEND	; done

KBD6:	mov TMP1,ITIME ; key still pressed
		and TMP1,TMP1
		brne	KBDEND	; time is not yet over
		ldi TMP1,ARPT2
		mov ITIME,TMP1	; prepare timer for autorepeat

KBDAK:	; perform actions
		lds TMP1,KEY_ST
		sbrs	TMP1,KEY0
		rjmp KBA1
		rcall FCN0
		rjmp KBDEND
KBA1:	sbrs	TMP1,KEY1
		rjmp KBA2
		rcall FCN1
		rjmp KBDEND
KBA2:	sbrs	TMP1,KEY2
		rjmp KBA3
		rcall FCN2
		rjmp KBDEND
KBA3:	sbrc	TMP1,KEY3
		rcall FCN3
KBDEND:

		and SCTR,SCTR	; when SCTR is 0 -> params to be saved into EEPROM
		brne SAVEND
		rcall SAVEPAR
		ldi TMP1,0xFF	; prevent repeated save (TIMER1 is stopped already)
		mov SCTR,TMP1
SAVEND:

		clr ACU1	; prepare for startup of next sine wave - phase 0, output voltage middle of range
		clr ACU2
		clr ACU3
		ldi TMP1,0x80
		out PORTD,TMP1

		ldi XH,HIGH(SINTBL)   ; MSB of SRAM sinus table
		cli		; no interrupt allowed during sine wave generation

;*********** SINE WAVE GENERATION, WHILE KEY IS DOWN ************
DLOOP:  ; when key is down, the program does nothing else but DDS signal generation! Full power needed.
		; one loop takes 10 clock cycles (loop freq 1.6 MHz) = 625 nsec
		sbic PINB,0    ; morse key connected to PB0. 0=KEY DOWN
		rjmp LOOP
		add ACU1,PHI1	; ACU += PHI
		adc ACU2,PHI2
		adc ACU3,PHI3	; X = pointer into SINTBL
		ld TMP1,X		; TMP1 = SINTBL[X]
		out PORTD,TMP1   ; output table value into R-2R network
		rjmp DLOOP

;======================================== end of main loop

; ******* SEND COMMAND OR DATA TO LCD *******
LCDC:	cbi	PORTC,0		; sends TMP1 to LCD as command (RS=0)
		rjmp	LCD1
LCDD:	sbi PORTC,0		; sends TMP1 to LCD as data (RS=1)
LCD1:	rcall	LCDS
		swap	TMP1
		rcall	LCDS
MOM:	ldi	TMP1,LCDWT	; wait 160 microseconds
LCD2:	push TMP1
		ldi TMP1,10
LCD3:	dec	TMP1
		brne LCD3
		pop	TMP1
		dec	TMP1
		brne	LCD2
		ret	

;****** sends upper 4 bits in TMP1 as a command ****
LCDX:	cbi	PORTC,0
;****** sends 4 upper bits TMP1 to LCD (TMP1 unchanged)	- data or command		
LCDS:	mov		TMP2,TMP1	
		andi	TMP2,0xF0
		lsr	TMP2
		lsr	TMP2		; move to position PC2 to PC5
		sbic PORTC,0	; read RS from port
		ori TMP2,1		; set bit0, when RS set (copy of RS bit)
		ori TMP2,2		; bit1 E = 1
		out	PORTC,TMP2 	; out data, RS and E
		nop
		nop
		nop
		nop
		nop
		nop
		nop				; pulse at E, wait a moment (more than 500 nsec)
		cbi	PORTC,1		; E = 0 - falling edge, write
		ret

;****** sends a standard text from EEPROM to LCD
TEXT:				; input: address of text in EEPROM in TMP1
		mov	TMP3,TMP1
		out	EEARL,TMP3
		ldi	TMP1,0
		out	EEARH,TMP1
		sbi	EECR,EERE
TEXT1:	sbic	EECR,EERE
		rjmp	TEXT1
		in	TMP1,EEDR	; text position byte, sent as command
		rcall	LCDC
TEXT4:	inc	TMP3
		out	EEARL,TMP3
		sbi	EECR,EERE
TEXT2:	sbic	EECR,EERE
		rjmp	TEXT2
		in	TMP1,EEDR	;ASCII character code 
		and	TMP1,TMP1
		breq	TEXT3	;0 ... string termination
		rcall	LCDD
		rjmp	TEXT4
TEXT3:	ret

;****** loop delay ******
WAIT:	ldi	TMP2,228	;timing procedure, value in TMP1 (x 100 microsec)
WAIT1:	dec	TMP2
		nop
		nop
		nop
		nop
		brne	WAIT1
		dec	TMP1
		brne	WAIT
		ret

;******** CALCULATION OF DDS PHASE INCREMENT FROM FREQ IN Hz ******
; PHI = 2^24 * FREQN / FRCLK
; takes 40 microseconds
DDSFR:
		lds XX1,FRQN
		lds XX2,FRQN+1
		lds XX3,FRQN+2	; XX = FREQ
		ldi TMP1,26	; initialize loop counter + 1
		mov CTR,TMP1
DF2:	dec CTR
		breq DF1

		lsl PHI1
		rol PHI2
		rol PHI3	; PHI shifted by 1 bit (multiplied by 2). Old content is rolled out.
		
		mov RES1,XX1	; LSB
		lds TMP1,FRCLK
		sub RES1,TMP1
		mov RES2,XX2
		lds TMP1,FRCLK+1
		sbc RES2,TMP1
		mov RES3,XX3
		lds TMP1,FRCLK+2
		sbc RES3,TMP1	; RES = XX - FRCLK
		brcs DF3		; jump when carry set (less than FRCLK, keep old value)
		mov XX1,RES1	; copy result into XX and increment PHI1 - only when XX >= FRCLK
		mov XX2,RES2
		mov XX3,RES3
		inc PHI1	; PHI += 1  - only when XX >= FRCLK
DF3:	lsl XX1
		rol XX2
		rol XX3		; XX *= 2
		rjmp DF2
DF1:	ret

;******* FUNCTION called by KEY 0 - cursor left (inc) **************
FCN0:
		lds TMP1,CPSN	; increment
		inc TMP1
		cpi TMP1,6
		brlt FC0A
		clr TMP1
FC0A:	sts	CPSN,TMP1
		rjmp PFRQ	; refresh display and return

;******* FUNCTION called by KEY 1 - Freq increment **************
FCN1:
		clr TMP3
		clr TMP2
		lds TMP1,CPSN
		cpi TMP1,0
		brne FC1A
		ldi TMP1,1		; step 1 Hz
		mov XX1,TMP1
		rjmp FC1X
FC1A:	cpi TMP1,1
		brne FC1B
		ldi TMP1, 10	; step 10 Hz
		mov XX1,TMP1
		rjmp FC1X
FC1B:	cpi TMP1,2
		brne FC1C
		ldi TMP1,100	; step 100 Hz
		rjmp FC1X
FC1C:	cpi TMP1,3
		brne FC1D
		ldi TMP1,0xE8
		ldi TMP2,3		; step 1000 Hz
		rjmp FC1X
FC1D:	cpi TMP1,4
		brne FC1E
		ldi TMP1,0x10	; step 10 kHz
		ldi TMP2,0x27
		rjmp FC1X
FC1E:	cpi TMP1,5
		brne FC1XX
		ldi TMP1,0xA0	; step 100 kHz
		ldi	TMP2,0x86
		ldi TMP3,1
FC1X:	mov XX1,TMP1
		mov XX2,TMP2
		mov XX3,TMP3
		rcall UFRQ		; perform freq calculation, range check
		rcall FQ2ASCII	; convert freq to decimal ASCII text
		rcall PFRQ		; print freq on  LCD
		rcall DDSFR		; calculates new phase increment
FC1XX:	ret

;******* FUNCTION called by KEY 2 - Freq decrement **************
; negative values added
FCN2:
		ldi TMP3,0xFF
		ldi TMP2,0xFF
		lds TMP1,CPSN
		cpi TMP1,0
		brne FC2A
		ldi TMP1,0xFF	; step 1 Hz
		mov XX1,TMP1
		rjmp FC2X
FC2A:	cpi TMP1,1
		brne FC2B
		ldi TMP1, 0xF6	; step 10 Hz
		mov XX1,TMP1
		rjmp FC2X
FC2B:	cpi TMP1,2
		brne FC2C
		ldi TMP1,0x9C	; step 100 Hz
		rjmp FC2X
FC2C:	cpi TMP1,3
		brne FC2D
		ldi TMP1,0x18
		ldi TMP2,0xFC		; step 1000 Hz
		rjmp FC2X
FC2D:	cpi TMP1,4
		brne FC2E
		ldi TMP1,0xF0	; step 10 kHz
		ldi TMP2,0xD8
		rjmp FC2X
FC2E:	cpi TMP1,5
		brne FC2XX
		ldi TMP1,0x60	; step 100 kHz
		ldi	TMP2,0x79
		ldi TMP3,0xFE
FC2X:	mov XX1,TMP1
		mov XX2,TMP2
		mov XX3,TMP3
		rcall UFRQ		; perform freq calculation, range check
		rcall FQ2ASCII	; convert freq to decimal ASCII text
		rcall PFRQ		; print freq on  LCD
		rcall DDSFR		; calculates new phase increment
FC2XX:	ret

;******* FUNCTION called by KEY 3 - cursor right (dec)**************
FCN3:
		lds TMP1,CPSN	; decrement
		and TMP1,TMP1
		breq FC3A
		dec TMP1
		rjmp FC3B
FC3A:	ldi TMP1,5
FC3B:	sts	CPSN,TMP1
		rjmp PFRQ	; refresh display and return


;****** FREQ to ASCII ******
; converts frequency in HEX into ASCII string
;******** PROCEDURE FREQUENCY TO ASCII ******
; convert frequency in Hz (up to 99.9.. MHz - 6 digits, 24 bit value) into ASCII chain
; input: FRQN binary freq
; result stored in SRAM buffer (fixed address)
; 1E5 = 0x186A0, 1E4 = 0x2710, 1E3 = 0x3E8, 100 = 0x64, 10 = 0xA

FQ2ASCII:
	ldi XL, LOW(FRQN)
	ldi XH, HIGH(FRQN)	; source
	ldi YL,LOW(BUF3B)
	ldi YH,HIGH(BUF3B)	; destination of copying
	clr TMP3
FQ2:
	cpi TMP3,3	; 3 bytes
	brge FQ1
	inc TMP3
	ld TMP1,X+
	st Y+,TMP1	; makes a copy of original value
	rjmp FQ2
FQ1:	; 1E5
	clr TMP2	; subtraction counter set to 0
	ldi YL,LOW(FRQA)
	ldi YH,HIGH(FRQA)	; where to store the result
FQ3:
	ldi XL,LOW(BUF3B)
	ldi XH,HIGH(BUF3B)	; point to LSB of copy
	ld XX1,X+
	ldi TMP1, 0xA0
	sub XX1,TMP1
	ld XX2,X+
	ldi TMP1, 0x86
	sbc XX2, TMP1
	ldi TMP1, 0x01
	ld XX3,X
	sbc XX3,TMP1	; subtract 1E5
	brcs FQ8 ; jumps when carry set (result is negative)
	inc TMP2
	ldi XL,LOW(BUF3B)
	ldi XH,HIGH(BUF3B)	; point to LSB of copy
	st X+,XX1
	st X+,XX2
	st X+,XX3
	rjmp FQ3
FQ8:
	ldi TMP1,0x30 ; ASCII '0'
	add TMP2,TMP1
	st Y+,TMP2	; store 1E5 digit

; 1E4
	clr TMP2	; subtraction counter set to 0
FQ9:
	ldi XL,LOW(BUF3B)
	ldi XH,HIGH(BUF3B)	; point to LSB of copy
	ld XX1,X+
	ldi TMP1, 0x10
	sub XX1,TMP1
	ld XX2,X+
	ldi TMP1, 0x27
	sbc XX2, TMP1
	clr TMP1
	ld XX3,X
	sbc XX3,TMP1	; subtract 1E4
	brcs FQA ; jumps when carry set (result is negative)
	inc TMP2
	ldi XL,LOW(BUF3B)
	ldi XH,HIGH(BUF3B)	; point to LSB of copy
	st X+,XX1
	st X+,XX2
	st X+,XX3
	rjmp FQ9
FQA:
	ldi TMP1,0x30 ; ASCII '0'
	add TMP2,TMP1
	st Y+,TMP2	; store 1E4 digit

; 1E3
	clr TMP2	; subtraction counter set to 0

FQB:
	ldi XL,LOW(BUF3B)
	ldi XH,HIGH(BUF3B)	; point to LSB of copy
	ld XX1,X+
	ldi TMP1, 0xE8
	sub XX1,TMP1
	ld XX2,X+
	ldi TMP1, 0x03
	sbc XX2, TMP1	; subtract 1E3
	brcs FQC ; jumps when carry set (result is negative)
	inc TMP2
	ldi XL,LOW(BUF3B)
	ldi XH,HIGH(BUF3B)	; point to LSB of copy
	st X+,XX1
	st X+,XX2
	rjmp FQB
FQC:
	ldi TMP1,0x30 ; ASCII '0'
	add TMP2,TMP1
	st Y+,TMP2	; store 1E3 digit

; 100
	clr TMP2	; subtraction counter set to 0
FQD:
	ldi XL,LOW(BUF3B)
	ldi XH,HIGH(BUF3B)	; point to LSB of copy
	ld XX1,X+
	ldi TMP1, 0x64
	sub XX1,TMP1
	ld XX2,X+
	clr TMP1
	sbc XX2, TMP1	; subtract 100
	brcs FQE ; jumps when carry set (result is negative)
	inc TMP2
	ldi XL,LOW(BUF3B)
	ldi XH,HIGH(BUF3B)	; point to LSB of copy
	st X+,XX1
	st X+,XX2
	rjmp FQD
FQE:
	ldi TMP1,0x30 ; ASCII '0'
	add TMP2,TMP1
	st Y+,TMP2	; store 100 digit

; 10
	clr TMP2	; subtraction counter set to 0
FQF:
	ldi XL,LOW(BUF3B)
	ldi XH,HIGH(BUF3B)	; point to LSB of copy
	ld XX1,X
	ldi TMP1, 0x0A
	sub XX1,TMP1	; subtract 10
	brcs FQG ; jumps when carry set (result is negative)
	inc TMP2
	st X,XX1
	rjmp FQF
FQG:
	ldi TMP1,0x30 ; ASCII '0'
	add TMP2,TMP1
	st Y+,TMP2	; store 10 digit

	ldi TMP1,0x30
	ld XX1,X
	add TMP1,XX1
	st Y,TMP1	; store 1 digit

	ret;

;***** UPDATE FREQ *****
; add freq increment (in XX1, XX2, XX3) to current freq
; check, whether freq in range ; if not, override the value
UFRQ:
		lds TMP1,FRQN
		add XX1,TMP1
		lds TMP1,FRQN+1
		adc	XX2,TMP1
		lds TMP1,FRQN+2
		adc	XX3,TMP1	; XX = calculated freq

		lds TMP1,FR1+2	; FR1 minimum freq
		cp	TMP1,XX3
		brlo UFM
		brne	UFL		; less - override LOW
		lds	TMP1,FR1+1	; equal - compare next byte
		cp	TMP1,XX2
		brlo UFM		; OK, calculated value accepted
		brne	UFL
		lds	TMP1,FR1	; equal, LSB byte
		cp TMP1,XX1
		brcc UFL

UFM:	lds TMP1,FR2+2	; FR2 max frequency
		cp	XX3,TMP1
		brlo UFOK
		brne UFH
		lds	TMP1,FR2+1
		cp	XX2,TMP1
		brlo UFOK
		brne UFH
		lds TMP1,FR2
		cp XX1, TMP1
		brcc UFH
		rjmp UFOK

UFL:	;lds XX1,FR1	; override with lower freq
		;lds XX2,FR1+1
		;lds XX3,FR1+2
		;rjmp UFOK
		rjmp UFEND		; no change
UFH:	;lds XX1,FR2	; override with higher freq
		;lds XX2,FR2+1
		;lds XX3,FR2+2
		rjmp UFEND		; no no change
UFOK:	sts FRQN,XX1
		sts FRQN+1,XX2
		sts FRQN+2,XX3
		rcall AST		; start/restart autosave counter
UFEND:	ret


;******* PRINT FREQUENCY - show freq on LCD display ****
;display FRQA buffer, insert dot for thousands Hz
;move cursor to required position
PFRQ:
		ldi TMP1,0xC5	; 2nd line, position 5
		rcall LCDC
		ldi	TMP3,7	; cycle counter (6 digits) + 1
		ldi	XH,HIGH(FRQA)
		ldi XL,LOW(FRQA)
PF1:	dec TMP3
		brne PF2
		rjmp PF4
PF2:	cpi TMP3,3	; 3rd position
		brne	PF3
		ldi TMP1,'.'
		rcall LCDD	; insert dot
PF3:	ld TMP1,X+
		rcall LCDD
		rjmp PF1
PF4:	lds TMP2,CPSN
		cpi TMP2,3
		brcs PF5
		inc TMP2
PF5:	ldi TMP1,0xCB	; position 2nd line, Hz digit
		sub TMP1,TMP2
		rjmp LCDC	; set cursor position and return

;****** START/RESTART AUTOSAVE TIMER ******
; register SCTR contains actual value, decremented every 4 seconds
; when reaches 0, autosave is performed in main loop (timer stops itself)
AST:
		cli
		ldi	TMP1,15	; 15x4 sec = 1 minute
		mov SCTR,TMP1
		ldi TMP1,5	; clock 1:1024 started
		out TCCR1B,TMP1
		sei
		ret

		
;****** INTERRUPT HANDLER OF TIMER 0 (8 bit) **********
; prescaler 1024, interrupt by 16 MHz clock every 16.394 msec
; decrements ITIME, till the value is zero
TIM0_OVF:
		in	ISREG,SREG
		and ITIME,ITIME
		breq TIM0_END	; is already 0, nothing to do
		dec ITIME
TIM0_END:	out	SREG,ISREG
		reti

;****** INTERRUPT HANDLER OF TIMER 1 (16 bit) **********
; prescaler 1024, interrupt by 16 MHz clock every 4.2 sec
; decrements ITIME, till the value is zero
TIM1_OVF:
		in	ISREG,SREG
		and SCTR,SCTR
		breq TIM1_S	; is already 0
		dec SCTR
		breq TIM1_S	; is just 0
		rjmp TIM1_END	; not yet 0, continue

TIM1_S:	clr ITMP1
		out TCCR1B,ITMP1	; stop TIMER1 (clock disconnected)

TIM1_END:	out	SREG,ISREG
		reti

;******** SAVE PERSISTANT PARAMETERS INTO EEPROM *****
; saves 3 bytes - frequency in Hz
; saved when 1 minute passed after last freq change
SAVEPAR:
		ldi XH,HIGH(FRQN)	; source SRAM
		ldi XL,LOW(FRQN)
		ldi YH,HIGH(EEFRQ)	; destination EEPROM
		ldi YL,LOW(EEFRQ)
		ldi TMP2,4		; 3 bytes +1
SPA2:	dec TMP2
		breq SPA1
		out EEARH,YH	; EEPROM address
		out EEARL,YL
		adiw Y,1		; Y += 1
		ld TMP1,X+		; read byte
		out EEDR,TMP1	; data register
		cli
			sbi EECR,EEMWE	; master write enable (must be set before write, it is autoreset by hardware, no IRQ before write command!)
			sbi EECR,EEWE	; write command
		sei
SPA3:	sbic EECR,EEWE	; loop till EEWE is cleard by hardware
		rjmp SPA3
		rjmp SPA2
SPA1:	ret	

;******** PROCEDURE reads one byte from EEPROM
RD_EE:	; input: address in Y register (R28,R29); output: byte in TMP1
	out	EEARL, YL
	out EEARH, YH
	sbi	EECR,EERE
RDEE1:	sbic	EECR,EERE
	rjmp	RDEE1
	in	TMP1,EEDR	; text position byte
	ret


;******** PROCEDURE reads block of data from EEPROM into SRAM ********
; input: X - pointer to SRAM, Y - pointer to EEPROM, TMP3 - number of bytes
EE_BLK:
	rcall	RD_EE
	st	X+,TMP1		; store to SRAM, increment pointers
	adiw Y,1
	dec	TMP3
	brne	EE_BLK
	ret
		
.ESEG
.ORG	0x0

;Each text even number of bytes (whole words), terminated by 0
VERS:	.DB		0x80,'L','F',' ','D','D','S',' ','1','.','0',' ','O','K','1','D','X',0
TXT1:	.DB		0xC0,'F','r','q',':',' ',0,0
TXT2:	.DB		0xCC,' ','k','H','z',0,0

EEFRQ:	// freq used during power up [Hz]
	.DB 0x40, 0xE2, 0x01	// 123 456 Hz = 0x01E240
//	.DB 0x1, 0x0, 0x0	// 1 Hz
EEFR1:	// minimum possible freq [Hz]
	.DB 0xE8, 0x03, 0x00	// 1 kHz = 0x3E8
EEFR2:	// maximum possible freq [Hz]
	.DB 0xC8, 0x02, 0x08	//525 kHz = 0x0802C8
EEFRCLK:	// DDS clock freq [Hz] - CPU clock / 10
	.DB 0x00, 0x6A, 0x18	// 1.6 MHz = 0x186A00

EESIN: // definition of 8 bit sin curve between 0 and 2*Pi
 .DB 128, 131, 134, 137, 140, 143, 146, 149, 152, 155 //idx0
 .DB 158, 162, 165, 167, 170, 173, 176, 179, 182, 185 //idx10
 .DB 188, 190, 193, 196, 198, 201, 203, 206, 208, 211 //idx20
 .DB 213, 215, 218, 220, 222, 224, 226, 228, 230, 232 //idx30
 .DB 234, 235, 237, 238, 240, 241, 243, 244, 245, 246 //idx40
 .DB 248, 249, 250, 250, 251, 252, 253, 253, 254, 254 //idx50
 .DB 254, 255, 255, 255, 255, 255, 255, 255, 254, 254 //idx60
 .DB 254, 253, 253, 252, 251, 250, 250, 249, 248, 246 //idx70
 .DB 245, 244, 243, 241, 240, 238, 237, 235, 234, 232 //idx80
 .DB 230, 228, 226, 224, 222, 220, 218, 215, 213, 211 //idx90
 .DB 208, 206, 203, 201, 198, 196, 193, 190, 188, 185 //idx100
 .DB 182, 179, 176, 173, 170, 167, 165, 162, 158, 155 //idx110
 .DB 152, 149, 146, 143, 140, 137, 134, 131, 128, 124 //idx120
 .DB 121, 118, 115, 112, 109, 106, 103, 100, 97,  93  //idx130
 .DB 90,  88,  85,  82,  79,  76,  73,  70,  67,  65  //idx140
 .DB 62,  59,  57,  54,  52,  49,  47,  44,  42,  40  //idx150
 .DB 37,  35,  33,  31,  29,  27,  25,  23,  21,  20  //idx160
 .DB 18,  17,  15,  14,  12,  11,  10,  9,   7,   6   //idx170
 .DB 5,   5,   4,   3,   2,   2,   1,   1,   1,   0   //idx180
 .DB 0,   0,   0,   0,   0,   0,   1,   1,   1,   2   //idx190
 .DB 2,   3,   4,   5,   5,   6,   7,   9,   10,  11  //idx200
 .DB 12,  14,  15,  17,  18,  20,  21,  23,  25,  27  //idx210
 .DB 29,  31,  33,  35,  37,  40,  42,  44,  47,  49  //idx220
 .DB 52,  54,  57,  59,  62,  65,  67,  70,  73,  76  //idx230
 .DB 79,  82,  85,  88,  90,  93,  97,  100, 103, 106 //idx240
 .DB 109, 112, 115, 118, 121, 124                     //idx250













